#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>
#include <errno.h>
#include "genbench.h"

#define N_OBJS		10000
#define DENSITY		0.10
#define Y_SPREAD	40.0
#define FLOOR_Y		-0.70
#define GRAVITY_Y	-9.81

#define TIME_STEP	0.04
#define RADIUS		0.12
#define DAMPING		0.20

#define DEBUG(s...)	fprintf(stderr, s)

#define HELP_MSG()	fprintf(stdout, \
"\nBenchmarking broad-phase algorithm \"%s\"\n\n" \
"Command line options [default]:\n\n" \
"   -n    Number of spheres to simulate [%d]\n" \
"   -r    Radius of spheres [%g]\n" \
"   -dt   Time step [%g]\n" \
"   -d    Portion of simulation space initially occupied by elements [%g]\n" \
"   -y    Initial vertical range [%g]\n" \
"   -f    Vertical lower constraint [%g]\n" \
"   -fd   Restitution coefficient of collisions on lower constraint [%g]\n" \
"   -g    Gravitational acceleration [%g]\n" \
"   -mid  Measure mid-phase time [disabled]\n" \
"   -hand Measure collision handling time [disabled]\n" \
"   -os   Output file for statistics [stdout]\n" \
"   -o    Output file for state [none]\n" \
"   -si   Stop after given iteration [never]\n" \
"\n", \
gb_stats.sim_name, \
N_OBJS, RADIUS, TIME_STEP, DENSITY, Y_SPREAD, FLOOR_Y, DAMPING, GRAVITY_Y)

/*
  Rationale on (not) using rdtsc
  
  The call to gettimeofday() implies switching to kernel mode, which is
  quite slow. This adds a significant bias on measure of short timings.
  There is an instruction (rdtsc) that returns a 64-bit timestamp counter
  incrementing at CPU frequency.
  But rdtsc has a few drawbacks: it is not portable to non-x86 architectures,
  and may give inconsistent results due to recent features such as
  frequency scaling, hyperthreading, and core switching.
*/

// *******************************************************************
// Internals
// *******************************************************************
#define GB_MAX_TIMERS	32
typedef struct gb_clock_t {
	struct timeval start;
	struct timeval total;
	int n_calls;
} gb_clock;


typedef struct gb_obj_t {
	double x, y, z;
	double vx, vy, vz;
	double radius;
	double r, g, b;
} gb_obj;

typedef struct gb_statistics_t {
	// simulation parameters
	int n_objs;
	double floor_y, floor_rest, gravity_y, time_step;
	int stop_iter, measure_mid, measure_hand;
	
	// simulation state
	char *sim_name;
	int frame;
	double x_min, x_max, y_min, y_max, z_min, z_max;
	
	// per-frame
	int pairs, false_pos;
	int bounces, nodes;
	size_t memory;
	
	// output
	FILE *stats_out;
	FILE *state_out;
} gb_statistics;


static gb_clock gb_timer[GB_MAX_TIMERS];
static gb_obj *gb_objs;
static gb_statistics gb_stats;

#ifndef rdtsc
#define rdtsc(low,high) \
     __asm__ __volatile__("rdtsc" : "=a" (low), "=d" (high))
#endif

// ***************************************************************************
// Utility random function
// ***************************************************************************
static double drand(double min, double max)
{
	double n01;
	
	n01 = random() / (1.0 + RAND_MAX);
	
	return(min + (n01 * (max - min)));
}

// ***************************************************************************
// Initialization
// ***************************************************************************
void gb_init(char *sim_name, int argc, char **argv)
{
	// parameters
	double density = DENSITY;
	double y_spread = Y_SPREAD;
	double radius = RADIUS;

	int t;
	double volume, xz_spread;
	
	// Command line arguments
	gb_stats.sim_name = sim_name;
	gb_stats.n_objs = N_OBJS;
	gb_stats.floor_y = FLOOR_Y;
	gb_stats.floor_rest = DAMPING;
	gb_stats.gravity_y = GRAVITY_Y;
	gb_stats.time_step = TIME_STEP;
	gb_stats.stats_out = stdout;
	gb_stats.state_out = NULL;
	gb_stats.stop_iter = -1;
	gb_stats.measure_mid = 0;
	gb_stats.measure_hand = 0;
	
	for (t = 1; t < argc; t++) {
		if ((!strcmp(argv[t], "--help")) || (!strcmp(argv[t], "-h"))) {
			HELP_MSG();
			exit(0);
		} else if (!strcmp(argv[t], "-mid")) {
			gb_stats.measure_mid = 1;
		} else if (!strcmp(argv[t], "-hand")) {
			gb_stats.measure_hand = 1;
		} else if (t + 1 >= argc) {
			HELP_MSG();
			exit(1);
		} else if (!strcmp(argv[t], "-n")) {
			gb_stats.n_objs = strtol(argv[++t], NULL, 0);
		} else if (!strcmp(argv[t], "-r")) {
			radius = strtod(argv[++t], NULL);
		} else if (!strcmp(argv[t], "-dt")) {
			gb_stats.time_step = strtod(argv[++t], NULL);
		} else if (!strcmp(argv[t], "-d")) {
			density = strtod(argv[++t], NULL);
		} else if (!strcmp(argv[t], "-y")) {
			y_spread = strtod(argv[++t], NULL);
		} else if (!strcmp(argv[t], "-f")) {
			gb_stats.floor_y = strtod(argv[++t], NULL);
		} else if (!strcmp(argv[t], "-fd")) {
			gb_stats.floor_rest = strtod(argv[++t], NULL);
		} else if (!strcmp(argv[t], "-g")) {
			gb_stats.gravity_y = strtod(argv[++t], NULL);
		} else if ((!strcmp(argv[t], "-os"))
		&& (!strcmp(argv[t+1], "-"))) {
			gb_stats.stats_out = stdout;
			t++;
		} else if (!strcmp(argv[t], "-os")) {
			gb_stats.stats_out = fopen(argv[++t], "w");
			if (!gb_stats.stats_out) {
				DEBUG("%s: %s\n", argv[t], strerror(errno));
				exit(1);
			}
		} else if ((!strcmp(argv[t], "-o"))
		&& (!strcmp(argv[t+1], "-"))) {
			if (gb_stats.stats_out == stdout)
				gb_stats.stats_out = stderr;
			gb_stats.state_out = stdout;
			t++;
		} else if (!strcmp(argv[t], "-o")) {
			gb_stats.state_out = fopen(argv[++t], "w");
			if (!gb_stats.state_out) {
				DEBUG("%s: %s\n", argv[t], strerror(errno));
				exit(1);
			}
		} else if (!strcmp(argv[t], "-si")) {
			gb_stats.stop_iter = strtol(argv[++t], NULL, 0);
		} else {
			HELP_MSG();
			exit(1);
		}
	}
	
	// Generate objects
	gb_objs = (gb_obj *)malloc(gb_stats.n_objs * sizeof(gb_obj));
	if (!gb_objs) {
		DEBUG("malloc(%d objs) failed\n", gb_stats.n_objs);
		exit(1);
	}
	
	volume = (4.0/3.0) * M_PI * pow(radius, 3.0);
	volume = volume * gb_stats.n_objs / density;
	xz_spread = sqrt(volume / y_spread);
	
	srandom(1);
	for (t = 0; t < gb_stats.n_objs; t++) {
		gb_objs[t].x = drand(-0.5 * xz_spread, 0.5 * xz_spread);
		gb_objs[t].y = drand(0.0, y_spread);
		gb_objs[t].z = drand(-0.5 * xz_spread, 0.5 * xz_spread);

		gb_objs[t].vx = gb_objs[t].vy = gb_objs[t].vz = 0;
		
		gb_objs[t].radius = radius;
		
		gb_objs[t].b = drand(0.0, 1.0);
		gb_objs[t].g = drand(0.0, 1.0) * gb_objs[t].b;
		gb_objs[t].r = drand(0.0, 1.0) * gb_objs[t].g;
	}
		
	gb_stats.x_min = gb_stats.z_min = -0.5 * xz_spread * 10.0;
	gb_stats.x_max = gb_stats.z_max =  0.5 * xz_spread * 10.0;
	gb_stats.y_min = ((gb_stats.floor_y < 0) ? gb_stats.floor_y : -1.0)
		* 10.0;
	gb_stats.y_max = y_spread * 1.1;
	
	gb_stats.frame = 0;

	gb_stats.pairs = gb_stats.false_pos = 0;
	gb_stats.bounces = 0;
	gb_stats.nodes = 0;
	gb_stats.memory = 0;
	
	gb_timers_init();

	// Initial distribution statistics
	if (gb_stats.stats_out) {
		fprintf(gb_stats.stats_out,
		"#                \"name\"   "
		" \"objs\" \"radius\" \"step\" \"density\""
		" \"y_range\" \"floor_y\" \"floor_rest\" \"grav\""
		" \"x0\"    \"x1\"    \"y0\"    \"y1\"    \"z0\"    \"z1\"\n"
		"# \"%s\" %7d %6.3f   %5.3f    %4.2f     %5.3f   "
		" %5.3f      %4.2f      %4.2f"
		"  %6.2f  %6.2f  %6.2f  %6.2f  %6.2f  %6.2f\n",
		gb_stats.sim_name, gb_stats.n_objs, radius, gb_stats.time_step,
		density, y_spread, gb_stats.floor_y, gb_stats.floor_rest,
		gb_stats.gravity_y,
		gb_min_x(), gb_max_x(),
		gb_min_y(), gb_max_y(),
		gb_min_z(), gb_max_z());
	}

	// Output initial distribution (frame -1)
	gb_dump_state(1);
}

// ***************************************************************************
// Memory stats:
// ***************************************************************************
size_t gb_malloc_total()
{
	return(gb_stats.n_objs * sizeof(gb_obj));
}

// ***************************************************************************
// Accessors
// ***************************************************************************
int  gb_nobj()
{
	return(gb_stats.n_objs);
}

void gb_obj_coord(int n, double *x, double *y, double *z)
{
	// if ((n < 0) || (n >= gb_stats.n_objs)) return;
	
	if (x) *x = gb_objs[n].x;
	if (y) *y = gb_objs[n].y;
	if (z) *z = gb_objs[n].z;
}

void gb_obj_color(int n, double *r, double *g, double *b)
{
	// if ((n < 0) || (n >= gb_stats.n_objs)) return;
	
	if (r) *r = gb_objs[n].r;
	if (g) *g = gb_objs[n].g;
	if (b) *b = gb_objs[n].b;
}

double gb_obj_radius(int n)
{
	// if ((n < 0) || (n >= gb_stats.n_objs)) return(0.0);

	return(gb_objs[n].radius);
}

void gb_tell_nodes(int nodes)
{
	gb_stats.nodes = nodes;
}

void gb_tell_memory(size_t mem)
{
	gb_stats.memory = mem;
}

double gb_min_x() { return(gb_stats.x_min); }
double gb_max_x() { return(gb_stats.x_max); }
double gb_min_y() { return(gb_stats.y_min); }
double gb_max_y() { return(gb_stats.y_max); }
double gb_min_z() { return(gb_stats.z_min); }
double gb_max_z() { return(gb_stats.z_max); }

// ***************************************************************************
// Collision narrow-phase handler
// ***************************************************************************
void gb_collision(int n1, int n2)
{
	double dx, dy, dz, d2, d, corr, mind;
	
	// if ((n < 0) || (n >= gb_stats.n_objs)) return;

	if (gb_stats.measure_mid)
		gb_timer_start(GB_MIDDLE_PHASE);
	
	gb_stats.pairs++;
	
	dx = gb_objs[n2].x - gb_objs[n1].x;
	dy = gb_objs[n2].y - gb_objs[n1].y;
	dz = gb_objs[n2].z - gb_objs[n1].z;
	
	d2 = (dx * dx) + (dy * dy) + (dz * dz);
	
	mind = gb_objs[n1].radius + gb_objs[n2].radius;
	if (d2 >= (mind*mind)) {
		// false positive
		gb_stats.false_pos++;
		if (gb_stats.measure_mid)
			gb_timer_stop(GB_MIDDLE_PHASE);
		return;
	}
	
	// true positive
	if (gb_stats.measure_hand)
		gb_timer_start(GB_COLLISION_HANDLING);
	
	d = sqrt(d2);
	if (d >= 0.0000001) {
		corr = 0.5001 * (mind - d) / d;
	} else {
		// DEBUG("ZERO DISTANCE\n");
		dx = 1.0;
		dy = dz = 0.0;
		corr = 0.5 * mind;
	}
	
	gb_objs[n2].x += corr * dx;
	gb_objs[n2].y += corr * dy;
	gb_objs[n2].z += corr * dz;
	gb_objs[n1].x -= corr * dx;
	gb_objs[n1].y -= corr * dy;
	gb_objs[n1].z -= corr * dz;

	if (gb_stats.measure_hand)
		gb_timer_stop(GB_COLLISION_HANDLING);
	if (gb_stats.measure_mid)
		gb_timer_stop(GB_MIDDLE_PHASE);
}


// ***************************************************************************
// Frame-to-frame transition
// ***************************************************************************
void gb_next_step()
{
	int t;
	
	// time integration
	for (t = 0; t < gb_stats.n_objs; t++) {
		gb_objs[t].vy += gb_stats.gravity_y * gb_stats.time_step;
		gb_objs[t].x += gb_objs[t].vx * gb_stats.time_step;
		gb_objs[t].y += gb_objs[t].vy * gb_stats.time_step;
		gb_objs[t].z += gb_objs[t].vz * gb_stats.time_step;
		
		if (gb_objs[t].y <= gb_stats.floor_y) {
			gb_objs[t].y = gb_stats.floor_y;
			if (gb_objs[t].vy < 0) {
				gb_objs[t].vy = -gb_stats.floor_rest
					* gb_objs[t].vy;
				gb_stats.bounces++;
			}
		}
	}
	
	// output
	gb_dump_state(0);
	gb_dump_stats(gb_stats.frame == 0);
	
	// exit
	if (gb_stats.frame == gb_stats.stop_iter) {
		exit(0);
	}
	
	// init next frame
	gb_stats.frame++;

	gb_stats.pairs = gb_stats.false_pos = 0;
	gb_stats.bounces = 0;
	gb_stats.nodes = 0;

	for (t = 0; t < GB_MAX_TIMERS; t++)
		gb_timer_clear(t);
}

// ***************************************************************************
// Stats Output
// ***************************************************************************
void gb_dump_stats(int with_header)
{
	int t;
	double mx, my, mz, vx, vy, vz, v;
	
	if (!gb_stats.stats_out) return;

	// mean
	mx = my = mz = 0.0;
	for (t = 0; t < gb_stats.n_objs; t++) {
		mx += gb_objs[t].x;
		my += gb_objs[t].y;
		mz += gb_objs[t].z;
	}
	mx /= (double)gb_stats.n_objs;
	my /= (double)gb_stats.n_objs;
	mz /= (double)gb_stats.n_objs;
	
	// std dev
	vx = vy = vz = 0.0;
	for (t = 0; t < gb_stats.n_objs; t++) {
		v = (gb_objs[t].x - mx); vx += v * v;
		v = (gb_objs[t].y - my); vy += v * v;
		v = (gb_objs[t].z - mz); vz += v * v;
	}
	vx = sqrt(vx / (double)gb_stats.n_objs);
	vy = sqrt(vy / (double)gb_stats.n_objs);
	vz = sqrt(vz / (double)gb_stats.n_objs);
	
	if (with_header) {
		fprintf(gb_stats.stats_out,
			"# \"Frame\" \"Objects\" \"mean x\" \"mean y\" "
			"\"mean z\" \"dev x\" \"dev y\" \"dev z\" \"bounces\" "
			"\"Coll.\" \"True+\"  \"Nodes\" \"Coll. Time\" "
			"\"Mid. Ph.\"  \"Handling\" \"Update Time\""
			" \"Memory\"\n");
	}
	
	fprintf(gb_stats.stats_out,
		" %6d %8d %8.3f %8.3f %8.3f %7.3f %7.3f %7.3f %8d",
		gb_stats.frame, gb_stats.n_objs, mx, my, mz, vx, vy, vz,
		gb_stats.bounces);
	
	fprintf(gb_stats.stats_out,
		" %8d %8d %9d %11.3f %11.3f %11.3f %11.3f %11zd\n",
		gb_stats.pairs, gb_stats.pairs - gb_stats.false_pos,
		gb_stats.nodes,
		gb_timer_ms(GB_COLLISION_PHASE),
		gb_stats.measure_mid ? gb_timer_ms(GB_MIDDLE_PHASE) : -1.0,
		gb_stats.measure_hand ? gb_timer_ms(GB_COLLISION_HANDLING)
			: -1.0,
		gb_timer_ms(GB_POSITION_UPDATE),
		gb_stats.memory);
	
	fflush(gb_stats.stats_out);
}

// ***************************************************************************
// State Output (this part is not portable)
// ***************************************************************************
void gb_dump_state(int with_header)
{
	int t;
	
	if (!gb_stats.state_out) return;

	if (with_header) {
		fwrite(&gb_stats.n_objs, 4, 1, gb_stats.state_out);

		for (t = 0; t < gb_stats.n_objs; t++) {
			fwrite(&gb_objs[t].radius, 8, 1, gb_stats.state_out);
			fwrite(&gb_objs[t].r, 8, 1, gb_stats.state_out);
			fwrite(&gb_objs[t].g, 8, 1, gb_stats.state_out);
			fwrite(&gb_objs[t].b, 8, 1, gb_stats.state_out);
		}
	}
	
	for (t = 0; t < gb_stats.n_objs; t++) {
		fwrite(&gb_objs[t].x, 8, 1, gb_stats.state_out);
		fwrite(&gb_objs[t].y, 8, 1, gb_stats.state_out);
		fwrite(&gb_objs[t].z, 8, 1, gb_stats.state_out);
	}
	
	fflush(gb_stats.state_out);
}

// ***************************************************************************
// Replay helper
// ***************************************************************************
int  gb_load_sim_file(FILE *f)
{
	int t;
	
	if (!f) {
		DEBUG("gl_load_sim_file(): null file\n");
		return(-1);
	}

	gb_stats.state_out = f;
	
	if (fread(&gb_stats.n_objs, 4, 1, gb_stats.state_out) != 1) {
		fclose(gb_stats.state_out);
		gb_stats.state_out = NULL;
		return(-1);
	}
	
	gb_objs = (gb_obj *)malloc(gb_stats.n_objs * sizeof(gb_obj));
	if (!gb_objs) {
		DEBUG("malloc(%d objs) failed\n", gb_stats.n_objs);
		fclose(gb_stats.state_out);
		gb_stats.state_out = NULL;
		return(-1);
	}
	
	for (t = 0; t < gb_stats.n_objs; t++) {
		if ((fread(&gb_objs[t].radius, 8, 1, gb_stats.state_out) != 1)
		||  (fread(&gb_objs[t].r, 8, 1, gb_stats.state_out) != 1)
		||  (fread(&gb_objs[t].g, 8, 1, gb_stats.state_out) != 1)
		||  (fread(&gb_objs[t].b, 8, 1, gb_stats.state_out) != 1)) {
			DEBUG("gb_load_sim_file(): %s\n", strerror(errno));
			fclose(gb_stats.state_out);
			gb_stats.state_out = NULL;
			return(-1);
		}
	}
	
	return(gb_load_frame());
}

int  gb_load_sim(char *file_name)
{
	FILE *f;

	f = fopen(file_name, "r");
	if (!f) {
		DEBUG("%s: %s\n", file_name, strerror(errno));
		return(-1);
	}

	return(gb_load_sim_file(f));
}

int  gb_load_frame()
{
	int t;
	
	if (!gb_stats.state_out) return(-1);
	
	for (t = 0; t < gb_stats.n_objs; t++) {
		if ((fread(&gb_objs[t].x, 8, 1, gb_stats.state_out) != 1)
		||  (fread(&gb_objs[t].y, 8, 1, gb_stats.state_out) != 1)
		||  (fread(&gb_objs[t].z, 8, 1, gb_stats.state_out) != 1)) {
			fclose(gb_stats.state_out);
			gb_stats.state_out = NULL;
			return(-1);
		}
	}
	
	return(0);
}

// ***************************************************************************
// Timers
// ***************************************************************************
void gb_timers_init()
{
	int t;
	
	for (t = 0; t < GB_MAX_TIMERS; t++)
		gb_timer_clear(t);
}

void gb_timer_clear(int id)
{
	// if ((id < 0) || (id >= GB_MAX_TIMERS)) return;
	
	gb_timer[id].total.tv_sec = 0;
	gb_timer[id].total.tv_usec = 0;
	gb_timer[id].n_calls = 0;
}


void gb_timer_start(int id)
{
	// if ((id < 0) || (id >= GB_MAX_TIMERS)) return;

	gettimeofday(&gb_timer[id].start, NULL);
}

void gb_timer_stop(int id)
{
	struct timeval now, diff;
	
	// if ((id < 0) || (id >= GB_MAX_TIMERS)) return;

	gettimeofday(&now, NULL);
	timersub(&now, &gb_timer[id].start, &diff);
	timeradd(&gb_timer[id].total, &diff, &gb_timer[id].total);
	gb_timer[id].n_calls++;
}

double gb_timer_ms(int id)
{
	// if ((id < 0) || (id >= GB_MAX_TIMERS)) return;

	return(1000.0 * gb_timer[id].total.tv_sec
		+ 0.001 * gb_timer[id].total.tv_usec);
}




