/*
    pool.c - linear tree and collision pool handling
    
    Copyright (C) 2009  Laurent Poirrier
    This file is part of the lin-kd-tree library.

    The lin-kd-tree library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, version 3 of the License.

    The lin-kd-tree library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "sort_pool.h"
#include "kd-tree.h"
#include "space.h"
#include "pool.h"

#define DEBUG(s...)		fprintf(stderr, s)

// ************************************************************************
// 
// ************************************************************************
kdtree_pool *kdtree_pool_new(KDTREE_INDEX nelements, KDTREE_PAIRS pairs_base)
{
	kdtree_pool *b;
	
	b = malloc(sizeof(kdtree_pool));
	if (!b) {
		DEBUG("kdtree_pool_new(): malloc(kdtree_pool) failed\n");
		return(NULL);
	}
	
	b->nel = 0;
	b->nel_alloc = nelements;
	b->nent = 0;
	
	b->elements = malloc(b->nel_alloc * sizeof(kdtree_element));
	if (!b->elements) {
		DEBUG("kdtree_pool_new(): malloc(kdtree_element) failed\n");
		free(b);
		return(NULL);
	}
	
	b->entries = malloc(b->nel_alloc * 8 * sizeof(kdtree_entry));
	if (!b->entries) {
		DEBUG("kdtree_pool_new(): malloc(kdtree_entry) failed\n");
		free(b->elements);
		free(b);
		return(NULL);
	}
	
	
	b->pair_gran = pairs_base;
	b->pair_buf = NULL;

	b->stats_out_of_bounds = 0;
	b->stats_tree_collisions = 0;
	b->stats_sphere_collisions = 0;
	b->stats_unique_collisions = 0;
	b->stats_malloc_total = sizeof(kdtree_pool)
		+ b->nel_alloc * sizeof(kdtree_element)
		+ b->nel_alloc * 8 * sizeof(kdtree_entry);
	return(b);
}

// ************************************************************************
// 
// ************************************************************************
void kdtree_pool_free(kdtree_pool *b)
{
	kdtree_pair_buffer *buf, *buf2;
	
	buf = b->pair_buf;
	while (buf) {
		buf2 = buf;
		buf = buf->next;
		free(buf2);
	}
	
	free(b->entries);
	free(b->elements);
	free(b);
}

// ************************************************************************
// 
// ************************************************************************
static kdtree_pair *kdtree_new_pair(kdtree_pool *b)
{
	kdtree_pair_buffer *buf;
	
	if ((!b->pair_buf) || (b->pair_buf->npairs == b->pair_gran)) {
		buf = malloc(sizeof(kdtree_pair_buffer) +
			sizeof(kdtree_pair) * b->pair_gran);
		
		if (!buf) {
			DEBUG("kdtree_new_pair(): malloc() failed\n");
			return(NULL);
		}
		
		buf->npairs = 0;
		buf->next = b->pair_buf;
		b->pair_buf = buf;
		
		b->stats_malloc_total += sizeof(kdtree_pair_buffer)
			+ sizeof(kdtree_pair) * b->pair_gran;
	} else {
		buf = b->pair_buf;
	}

	return(&buf->pairs[buf->npairs++]);
}


// ************************************************************************
// 
// ************************************************************************
int  kdtree_pool_add_sphere(kdtree_pool *b, const kdtree_sphere *bs, void *obj)
{
	kdtree_element *e;
	
	if (b->nel >= b->nel_alloc) {
		DEBUG("kdtree_pool_add(): too many elements\n");
		return(-1);
	}
	
	e = &b->elements[b->nel++];
	
	e->pointer = obj;
	memcpy(&e->sphere, bs, sizeof(kdtree_sphere));
	e->first = NULL;
	
	return(0);
}

// ************************************************************************
// 
// ************************************************************************
void kdtree_pool_build_tree(kdtree_pool *b, const kdtree_space *s)
{
	KDTREE_INDEX i, j;
	kdtree_iaabb ibb;
	kdtree_node xk[8];
	int nx;
	
	for (i = 0; i < b->nel; i++) {
		if (kdtree_pack_sphere(s, &ibb, &b->elements[i].sphere)) {
			kdtree_root(xk);
			nx = 1;
			b->stats_out_of_bounds++;
		} else {
			nx = kdtree_split_place(xk, &ibb, 1, 1, 1);
		}
	
		for (j = 0; j < nx; j++) {
			memcpy(&b->entries[b->nent].path, &xk[j],
				sizeof(kdtree_node));
#ifdef KDTREE_INDEXED_ENTRIES
			b->entries[b->nent].idx = i;
#else
			b->entries[b->nent].element = &b->elements[i];
#endif
			b->nent++;
		}
	}
}

// ************************************************************************
// 
// ************************************************************************
void kdtree_pool_sort_tree(kdtree_pool *b)
{
	kdtree_pool_quicksort(b->entries, 0, b->nent - 1);
}

// ************************************************************************
// 
// ************************************************************************
static void kdtree_pool_interact(kdtree_pool *b,
#ifdef KDTREE_INDEXED_PAIRS
	KDTREE_INDEX idx_i, KDTREE_INDEX idx_j,
#endif
	kdtree_element *ei, kdtree_element *ej)
{
	KDTREE_SCALAR dx, dy, dz, dr2, dmax2;
	kdtree_pair *p;
	
	b->stats_tree_collisions++;

	// Bounding sphere culling
	dx = ei->sphere.x - ej->sphere.x;
	dy = ei->sphere.y - ej->sphere.y;
	dz = ei->sphere.z - ej->sphere.z;
	dr2 = (dx * dx) + (dy * dy) + (dz * dz);
	
	dmax2 = ei->sphere.r + ej->sphere.r;
	dmax2 *= dmax2;
	
	if (dr2 > dmax2)
		return;

	b->stats_sphere_collisions++;
	
	// Duplicate interaction culling
	p = ei->first;
	
	while (p) {
#ifdef KDTREE_INDEXED_PAIRS
		if (p->idx_A == idx_i) {
			if (p->idx_B == idx_j) return;
			
			p = p->next_A;
		} else {
			if (p->idx_A == idx_j) return;
			
			p = p->next_B;
		}
#else
		if (p->element_A == ei) {
			if (p->element_B == ej) return;
			
			p = p->next_A;
		} else {
			if (p->element_A == ej) return;
			
			p = p->next_B;
		}
#endif
	}

	b->stats_unique_collisions++;
	
	// New interaction pair
	p = kdtree_new_pair(b);
	if (!p) return;
	
#ifdef KDTREE_INDEXED_PAIRS
	p->idx_A = idx_i;
	p->idx_B = idx_j;
#else
	p->element_A = ei;
	p->element_B = ej;
#endif

	p->next_A = ei->first;
	p->next_B = ej->first;
	ei->first = p;
	ej->first = p;
	return;
}

void kdtree_pool_sweep_tree(kdtree_pool *b)
{
	kdtree_entry *ip, *jp;
	KDTREE_ENTRIES i, j;
	
	i = 0;
	ip = b->entries;
	while (i < b->nent) {
		j = i + 1;
		jp = ip + 1;
		
		while ((j < b->nent)
		&&  (KDTREE_INTERSECT(ip->path, jp->path))) {
#ifdef KDTREE_INDEXED_PAIRS
			kdtree_pool_interact(b, ip->idx, jp->idx,
				&b->elements[ip->idx], &b->elements[jp->idx]);
#else
#ifdef KDTREE_INDEXED_ENTRIES
			kdtree_pool_interact(b,
				&b->elements[ip->idx], &b->elements[jp->idx]);
#else
			kdtree_pool_interact(b, ip->element, jp->element);
#endif
#endif
			jp++;
			j++;
		}
		
		ip++;
		i++;
	}
}


